"""
This is a rewrite of the HTTP transport to use httptools parser
It currently still uses h11 for client transport
"""

import asyncio
import traceback
import time

import h11
try:
    from httptools import HttpRequestParser, HttpResponseParser, HttpParserError, \
        parse_url
except ImportError:
    pass  # You need to have httptools installed to use this

from ..webtypes import Request, Response
from .transportabc import TransportABC, ClientTransportABC


class ClosedError(Exception):
    """ Error for closed connections """


class _HTTPClientConnection:
    slots = ('reader', 'writer', 'conn', 'port', 'service')

    def __init__(self, service, port):
        self.reader = None
        self.writer = None
        self.port = port
        self.service = service
        self.conn = h11.Connection(our_role=h11.CLIENT)

    async def connect(self):
        self.reader, self.writer = await \
            asyncio.open_connection(self.service, self.port)

    def send(self, event):
        data = self.conn.send(event)
        if data is not None:
            self.writer.write(data)

    async def next_event(self):
        while True:
            event = self.conn.next_event()
            if event is h11.NEED_DATA:
                data = await self.reader.read(1064)
                self.conn.receive_data(data)

            else:
                return event

    def close(self):
        self.send(h11.ConnectionClosed())
        self.writer.close()


class HTTPClientTransport(ClientTransportABC):
    """Client implementation of the HTTP transport protocol"""
    def _get_connection_for_service(self, service):
        pass

    async def make_request(self, service, method, path, body=None, query=None,
                           headers=None, correlation_id=None,
                           content_type=None, port=80, **kwargs):
        # form request object
        path = path.replace('.', '/')
        if not path.startswith('/'):
            path = '/' + path
        if headers is None:
            headers = {}
        headers['Host'] = service
        if port != 80:
            headers['Host'] += ':{}'.format(port)
        headers['Connection'] = 'close'
        if correlation_id:
            headers['X-Correlation-Id'] = correlation_id
        if query:
            path += '?' + query
        if content_type:
            headers['Content-Type'] = content_type
        if body:
            headers['Content-Length'] = str(len(body))

        # now make a connection and send it
        connection = _HTTPClientConnection(service, port)
        await connection.connect()
        connection.send(h11.Request(method=method, target=path,
                                    headers=headers.items()))
        if body:
            connection.send(h11.Data(data=body))

        connection.send(h11.EndOfMessage())

        response = await connection.next_event()
        assert type(response) is h11.Response

        # form response object

        status_code = response.status_code
        response_headers = response.headers
        response_headers = {k.decode(): v.decode() for k,v in response_headers}

        result = Response(headers=response_headers,
                          correlation_id=correlation_id,
                          status=status_code)

        body = b''
        event = None
        while type(event) is not h11.EndOfMessage:
            event = await connection.next_event()
            if type(event) is h11.Data:
                body += body

        result._data = body.decode()
        connection.close()
        return result


class HTTPTransport(TransportABC):
    """ Server implementation of the HTTP transport protocol"""

    def get_client(self):
        return HTTPClientTransport()

    def __init__(self, port=8080, prefix=None, shutdown_grace_period=5,
                 shutdown_wait_period=1):
        """
         HTTP Transport for listening on http
         :param port: The port to lisen on (0.0.0.0 will always be used)
         :param prefix: the path prefix to remove from all url's
         :param shutdown_grace_period: Time to wait for server to shutdown
         before connections get forceably closed. The only way for connections
         to not be forcibly closed is to have some connection draining in front
         of the service for deploys. Most docker schedulers will do this for you.
         :param shutdown_wait_period: Time to wait after recieving the sigterm
         before starting shutdown 
         """
        self.port = port
        if prefix is None:
            prefix = ''
        self.prefix = prefix
        self._handler = None
        self._server = None
        self._loop = None
        self._done_future = asyncio.Future()
        self._connections = set()
        self.shutdown_grace_period = shutdown_grace_period
        self.shutdown_wait_period = shutdown_wait_period

    def listen(self, *, loop: asyncio.AbstractEventLoop):
        self._loop = loop

    async def start(self, request_handler):
        self._handler = request_handler
        self._server = await self._loop.create_server(
            lambda: _HTTPServerProtocol(parent=self, loop=self._loop),
            host='0.0.0.0',
            port=self.port,
            reuse_address=True,
            reuse_port=True)
        print('-- Listening for HTTP on port {} --'.format(self.port))
        try:
            await self._done_future
        except asyncio.CancelledError:
            pass
        await asyncio.sleep(self.shutdown_wait_period)
        # wait for connections to stop
        times_no_connections = 0
        for i in range(self.shutdown_grace_period):
            if not self._connections:
                times_no_connections += 1
            else:
                times_no_connections = 0
                for con in self._connections:
                    con.attempt_close()

            if times_no_connections > 3:
                # three seconds with no connections
                break
            await asyncio.sleep(1)

        # Shut the server down
        self._server.close()
        await self._server.wait_closed()

    async def handle_incoming_request(self, request):
        response = await self._handler(request)
        return response

    def shutdown(self):
        self._done_future.cancel()


class _HTTPServerProtocol(asyncio.Protocol):
    """ HTTP Protocol handler.
        Should only be used by HTTPServerTransport
    """
    __slots__ = ('_parent', '_transport', 'data', 'http_parser',
                 'request')

    def __init__(self, *, parent, loop):
        self._parent = parent
        self._transport = None
        self.data = None
        self.http_parser = HttpRequestParser(self)
        self.request = None
        self._loop = loop

    """ The next 3 methods are for asyncio.Protocol handling """
    def connection_made(self, transport):
        self._transport = transport
        self._parent._connections.add(self)

    def connection_lost(self, exc):
        self._parent._connections.discard(self)

    def data_received(self, data):
        try:
            self.http_parser.feed_data(data)
        except HttpParserError as e:
            traceback.print_exc()
            print(self.request.__dict__)
            self.send_response(Response(status=400,
                                        body={'reason': 'Invalid HTTP'}))

    """ 
    The following methods are for HTTP parsing (from httptools)
    """
    def on_message_begin(self):
        self.request = Request()
        self.data = b''

    def on_header(self, name, value):
        key = name.decode('ASCII').lower()
        val = value.decode()
        self.request.headers[key] = val
        if key == 'x-correlation-id':
            self.request.correlation_id = val
        if key == 'content-type':
            self.request.content_type = val

    def on_headers_complete(self):
        self.request.method = self.http_parser.get_method().decode('ASCII')

    def on_body(self, body: bytes):
        self.data += body

    def on_message_complete(self):
        self.request.body = self.data
        task = self._loop.create_task(
            self._parent.handle_incoming_request(self.request)
        )
        task.add_done_callback(self.handle_response)

    def on_url(self, url):
        url = parse_url(url)
        if url.query:
            self.request.query_string = url.query.decode('ASCII')
        self.request.path = url.path.decode('ASCII')

    """
    End parsing methods
    """

    def handle_response(self, future):
        try:
            self.send_response(future.result())
        except Exception:
            traceback.print_exc()
            self.send_response(
                Response(status=500,
                         body={'reason': 'Something really bad happened'}))

    def send_response(self, response):
        headers = '''\
HTTP/1.1 {status_code} {status_message}\r
Connection: keep-alive\r
'''.format(status_code=response.status.value,
           status_message=response.status.phrase,
           )
        if response.data:
            headers += 'Content-Type: {}\r\n'.format(response.content_type)
            headers += 'Content-Length: {}\r\n'.format(len(response.data))
            if ('transfer-encoding' in response.headers or
                        'Transfer-Encoding' in response.headers):
                print('Httptoolstransport currently doesnt support '
                      'chunked mode, attempting without.')
                response.headers.pop('transfer-encoding', None)
                response.headers.pop('Transfer-Encoding', None)
        else:
            headers += 'Content-Length: {}\r\n'.format(0)
        for header, value in response.headers.items():
            headers += '{header}: {value}\r\n'.format(header=header,
                                                      value=value)

        result = headers.encode('ASCII') + b'\r\n'
        if response.data:
            result += response.data

        self._transport.write(result)
        self.request = 0
        self.data = 0

    def attempt_close(self):
        if self.request == 0:
            self._transport.close()



from .httptransport import HTTPTransport, HTTPClientTransport
from .httptoolstransport import HTTPTransport as HTTPToolsTransport
from .rabbitmqtransport import RabbitMQTransport
from .transportabc import TransportABC
from .testtransport import TestTransport





import asyncio
import uuid

import aioamqp
from aioamqp.channel import Channel

from .transportabc import TransportABC, ClientTransportABC
from ..webtypes import Request, Response


class RabbitMQClientTransport(ClientTransportABC):
    def __init__(self, channel):
        self.channel = channel
        self._response_futures = {}
        self.response_queue_name = str(uuid.uuid1()).encode()
        self._consumer_tag = None

    async def make_request(self, service: str, method: str, path: str,
                           body: bytes = None, query: str = None,
                           headers: dict = None, correlation_id: str = None,
                           content_type: str = None, **kwargs):

        if not self._consumer_tag:
            await self.start()
        path = path.replace('/', '.').lstrip('.')
        if headers is None:
            headers = {}
        if query:
            headers['x-wasp-query-string'] = query

        if not body:
            body = b'None'
        message_id = str(uuid.uuid4())
        properties = {
            'headers': headers,
            'reply_to': self.response_queue_name,
            'correlation_id': correlation_id,
            'message_id': message_id,
            'expiration': '30000',
            'type': method,
            'app_id': 'hello',
        }
        if content_type:
            properties['content_type'] = content_type

        await self.channel.basic_publish(exchange_name='amq.topic',
                                         routing_key=path,
                                         properties=properties,
                                         payload=body)

        future = asyncio.Future()
        self._response_futures[message_id] = future
        return await future

    async def start(self):
        await self.channel.queue_declare(queue_name=self.response_queue_name,
                                         durable=False,
                                         exclusive=True,
                                         auto_delete=True)
        self._consumer_tag = (await self.channel.basic_consume(
            self.handle_responses,
            queue_name=self.response_queue_name,
            no_ack=True)).get('consumer_tag')

    async def handle_responses(self, channel, body, envelope, properties):
        future = self._response_futures[properties.message_id]

        headers = properties.headers
        status = headers.pop('Status')

        response = Response(headers=headers,
                            correlation_id=properties.correlation_id,
                            body=body,
                            status=status,
                            content_type=properties.content_type)

        future.set_result(response)


class RabbitMQTransport(TransportABC):
    def __init__(self, *, url, port=5672, queue='', virtualhost='/',
                 username='guest', password='guest',
                 ssl=False, verify_ssl=True):
        self.host = url
        self.port = port
        self.virtualhost = virtualhost
        self.queue = queue
        self.username = username
        self.password = password
        self.ssl = ssl
        self.verify_ssl = verify_ssl
        self._transport = None
        self._protocol = None
        self.channel = None
        self._app = None
        self._loop = None
        self._consumer_tag = None
        self._counter = 0
        self._handler = None
        self._done_future = asyncio.Future()

    def get_client(self):
        return RabbitMQClientTransport(self.channel)

    async def declare_exchange(self):
        pass

    async def declare_queue(self):
        pass

    async def bind_to_exchange(self, *, exchange, routing_key):
        await self.channel.queue_bind(exchange_name=exchange,
                                      queue_name=self.queue,
                                      routing_key=routing_key)

    async def start(self, handler):
        self._handler = handler
        # Need to reconnect because of potential forking affects
        await self.close()
        await self.connect()
        self._consumer_tag = (await self.channel.basic_consume(
            self.handle_request,
            queue_name=self.queue,
            no_ack=True)).get('consumer_tag')

        try:
            await self._done_future
        except asyncio.CancelledError:
            pass

        print('shutting down rabbit')
        # shutting down
        await self.channel.basic_cancel(self._consumer_tag)

        while self._counter > 0:
            await asyncio.sleep(1)

    def listen(self, *, loop):
        async def setup():

            await self.connect(loop=loop)
            await self.channel.queue_declare(queue_name=self.queue)
            await self.channel.basic_qos(prefetch_count=1, prefetch_size=0)

        loop.run_until_complete(setup())
        print('-- Listening for rabbitmq messages on queue {} --'
              .format(self.queue))

    async def connect(self, loop=None):
        self._transport, self._protocol = await aioamqp.connect(
            host=self.host,
            port=self.port,
            virtualhost=self.virtualhost,
            login=self.username,
            password=self.password,
            ssl=self.ssl,
            verify_ssl=self.verify_ssl,
            loop=loop
        )
        self.channel = await self._protocol.channel()

    async def close(self):
        await self._protocol.close()
        self._transport.close()

    async def handle_request(self, channel: Channel, body, envelope,
                             properties, futurize=True):
        """
        the 'f' param is simply because aioamqp doesnt send another job until
         this method returns (completes), so we ensure the future of
         ourselves and return immediately so we can handle many requests
         at a time.
        """
        if futurize:
            asyncio.ensure_future(
                self.handle_request(channel, body, envelope, properties,
                                    futurize=False))
            return

        self._counter += 1
        route = envelope.routing_key
        print(route)
        headers = properties.headers or {}
        query = headers.pop('x-wasp-query-string', '').lstrip('?')
        headers['content-type'] = properties.content_type
        headers['content-encoding'] = properties.content_encoding
        correlation_id = properties.correlation_id
        message_id = properties.message_id
        reply_to = properties.reply_to
        method = properties.type

        request = Request(
            headers=headers,
            path=route,
            correlation_id=correlation_id,
            method=method,
            query_string=query,
            body=body,
        )

        response = await self._handler(request)
        if not reply_to:
            # nowhere to reply. No point it trying to send a response
            return

        response.headers['Status'] = str(response.status.value)

        payload = response.data or b'None'

        properties = {
            'correlation_id': response.correlation_id,
            'headers': response.headers,
            'content_type': response.content_type,
            'message_id': message_id,
            'expiration': '30000',
        }
        await channel.basic_publish(exchange_name='',
                                    payload=payload,
                                    routing_key=reply_to,
                                    properties=properties)
        self._counter -= 1

    def shutdown(self):
        print('Rabbit got shutdown signal')
        self._done_future.cancel()






from enum import Enum

from waspy import webtypes

"""
The below constant is special key in the router dictionary that determines an
id section of the url. The / character is the only character that
can't possibly be used in a url path or id, since it is the path delimiter.
Therefore, there are '/' characters in the key, so that it can never 
accidentally be overridden.

p.s. '_id' is to make the dictionary slightly more readable when debugging
"""
ID_KEY = '/_id/'


class NotAValidURLError(Exception):
    """ When a url path syntax is not valid """


class Methods(Enum):
    GET = 'GET'
    POST = 'POST'
    PUT = 'PUT'
    DELETE = 'DELETE'
    OPTIONS = 'OPTIONS'
    PATCH = 'PATCH'
    HEAD = 'HEAD'

async def _send_404(request):
    raise webtypes.ResponseError(status=404)


class Router:
    def __init__(self):
        self._routes = {}
        """
        Routes looks like:
        {path_section1: {path_section2: {method: (handler, params)}}}
                before startup and
        {path_section1: {papath_section2 {method: (wrapped, handler, params)}}
                after startup
        """
        self._static_routes = {}
        """
        Static routes are just simple d[url][method] lookups and skip
        middlewares
        """
        self.options_handler = None

        self.handle_404 = _send_404  # the 404 route will skip middlewares

    def _get_and_wrap_routes(self, _d=None):
        if _d is None:
            _d = self._routes
        for key, value in _d.items():
            # if value is a dictionary, keep going
            # if value is a tuple, then wrap it!
            if isinstance(value, dict):
                yield from self._get_and_wrap_routes(_d=value)
            else:
                handler, params = value
                wrapped = yield handler
                _d[key] = (wrapped, handler, params)

    def get_handler_for_request(self, request):
        method = request.method
        path = request.path
        route = path.lstrip('/').replace('/', '.')
        if method == Methods.OPTIONS and self.options_handler is not None:
            # Is this an OPTION and do we have a generic options handler?
            request._handler = self.options_handler
            return self.options_handler

        try:
            return self._static_routes[route][method]
        except KeyError:
            # not in static routes
            pass

        d = self._routes
        params = []
        try:
            for portion in route.split('.'):
                sub = d.get(portion, None)
                if sub is None:  # must be an ID field
                    key = ID_KEY
                    param = portion
                    if ':' in portion:
                        param, action = portion.split(':', 1)
                        key += ':' + action

                    sub = d[key]
                    params.append(param)
                d = sub
            wrapped, handler, keys = d[method]
        except KeyError:
            # No handler for given route
            request._handler = self.handle_404
            return self.handle_404
        assert len(keys) == len(params)
        for key, param in zip(keys, params):
            request.path_params[key] = param
        request._handler = handler
        return wrapped

    def add_static_route(self, method: str, route: str, handler: callable,
                         skip_middleware=False):
        """
        Adds a static route. A static route is a special route that
        doesnt follow any of the normal rules, and never has any path
        parameters.
        Ideally, this is used for non-public facing endpoints such as
        "/healthcheck", or "/stats" or something of that nature.

        All static routes SKIP middlewares
        """
        if not isinstance(method, Methods):
            method = Methods(method.upper())
        route = route.lstrip('/').replace('/', '.')
        if route not in self._static_routes:
            self._static_routes[route] = {}
        self._static_routes[route][method] = handler

    def add_route(self, method: str, route: str, handler: callable):
        if not isinstance(method, Methods):
            method = Methods(method.upper())

        route = route.replace('/', '.').lstrip('.')
        d = self._routes
        params = []
        for portion in route.split('.'):
            if portion.startswith('{') and '}' in portion:
                sections = portion.lstrip('{').split('}', maxsplit=1)
                params.append(sections[0])
                key = ID_KEY
                if sections[1]:
                    if not sections[1].startswith(':'):
                        raise NotAValidURLError(
                            'Cant have an id mixed with '
                            'a static word without a colon')
                    key += sections[1]

            else:
                key = portion
            if key not in d:
                d[key] = {}
            d = d[key]

        d[method] = handler, params

    def add_get(self, route: str, handler: callable):
        self.add_route(Methods.GET, route, handler)

    def add_post(self, route: str, handler: callable):
        self.add_route(Methods.POST, route, handler)

    def add_put(self, route: str, handler: callable):
        self.add_route(Methods.PUT, route, handler)

    def add_delete(self, route: str, handler: callable):
        self.add_route(Methods.DELETE, route, handler)

    def add_patch(self, route: str, handler: callable):
        self.add_route(Methods.PATCH, route, handler)

    def add_head(self, route: str, handler: callable):
        self.add_route(Methods.HEAD, route, handler)

    def add_options(self, route: str, handler: callable):
        self.add_route(Methods.OPTIONS, route, handler)

    def add_generic_options_handler(self, handler: callable):
        """
        Add a handler for all options requests. This WILL bypass middlewares
        """
        self.options_handler = handler



import json
from urllib import parse
import uuid

import asyncio

from .webtypes import QueryParams, Request, Methods


class Client:
    """ Generic Client class for making a wasp client """
    __slots__ = ('transport',)

    def __init__(self, transport=None, **kwargs):
        if not transport:
            from waspy.transports import HTTPClientTransport
            transport = HTTPClientTransport(**kwargs)
        self.transport = transport

    def make_request(self, method, service, path, body=None,
                           query_params: QueryParams=None,
                           headers: dict=None,
                           correlation_id: str=None,
                           content_type: str='application/json',
                           context: Request=None,
                           **kwargs) -> asyncio.coroutine:
        """
        Make a request to another service. If `context` is provided, then
        context and correlation will be pulled from the provided request
        object for you. This includes credentials, correlationid,
        and service-headers.

        :param method: GET/PUT/PATCH, etc.
        :param service: name of service
        :param path: request object path
        :param body: body of request
        :param query_params:
        :param headers:
        :param correlation_id:
        :param content_type:
        :param context: A request object from which a "child-request"
            will be made
        :param kwargs: Just a place holder so transport specific options
            can be passed through
        :return:
        """
        if not isinstance(method, Methods):
            method = Methods(method.upper())
        if content_type == 'application/json' and isinstance(body, dict):
            body = json.dumps(body)
        if isinstance(query_params, dict):
            query_string = parse.urlencode(query_params)
        elif isinstance(query_params, QueryParams):
            query_string = str(QueryParams)
        else:
            query_string = ''

        if context:
            if not correlation_id:
                correlation_id = context.correlation_id

        if not correlation_id:
            correlation_id = str(uuid.uuid4())
        if isinstance(body, str):
            body = body.encode()
        response = self.transport.make_request(
            service, method.name, path, body=body, query=query_string,
            headers=headers, correlation_id=correlation_id,
            content_type=content_type, **kwargs)
        return response  # response is a coroutine that must be awaited

    def get(self, service, path, **kwargs):
        """ Make a get request (this returns a coroutine)"""
        return self.make_request(Methods.GET, service, path, **kwargs)

    def post(self, service, path, body, **kwargs):
        """ Make a post request (this returns a coroutine)"""
        return self.make_request(Methods.POST, service, path, body=body,
                                 **kwargs)

    def put(self, service, path, body, **kwargs):
        """ Make a put request (this returns a coroutine)"""
        return self.make_request(Methods.POST, service, path, body=body,
                                 **kwargs)

    def patch(self, service, path, body, **kwargs):
        """ Make a patche requests (this returns a coroutine)"""
        return self.make_request(Methods.PATCH, service, path, body=body,
                                 **kwargs)

    def delete(self, service, path, **kwargs):
        """ Make a delete requests (this returns a coroutine)"""
        return self.make_request(Methods.DELETE, service, path, **kwargs)
